unit uWaveIO;

interface

uses
   Windows, Classes, MMSystem,
   uInterfaces;

const
  cnHdrCount = 4;
  cnBufTime = 250; // ms

type
  TWAVEFORMATEXTENSIBLE = packed record
    Format: TWaveFormatEx;
    Samples: Word; // wValidBitsPerSample, wSamplesPerBlock, wReserved
    ChannelMask: DWORD;
    SubFormat: TGUID;
  end;

  TWaveThread = class;

  TWaveControl = class(TInterfacedObject)
  private
    FThread: TWaveThread;
    FSetup: IWaveSetup;
    FFormatExt: TWAVEFORMATEXTENSIBLE;
    FBuffers: Pointer;
    FWHdrs: array [0..cnHdrCount - 1] of record
      Hdr: TWaveHdr;
      Spl: LongWord;
    end;
    procedure GetFormat;
    procedure AllocBuffers(aCount: Integer);
    function GetChannels: Integer;
  protected
    FBufData: array of LongInt;
    procedure Synchronize(Method: TThreadMethod);
    procedure SetChannels(aValue: Integer); virtual;
  public
    constructor Create(aSetup: IWaveSetup);
    destructor Destroy; override;
    procedure Stop;
    function Running: Boolean;
    property Channels: Integer read GetChannels write SetChannels;
  end;

  TWaveThread = class(TThread)
  private
    FControl: TWaveControl;
    FEvent: THandle;
    FDeviceHD: Integer;
    FWHdrID: LongWord;
    FpWHdr: PWaveHdr;
    FSamples: LongWord;
    FCount: LongWord;
    procedure NextWHdr;
  public
    constructor Create(aControl: TWaveControl);
    destructor Destroy; override;
  end;

  TWaveIn = class(TWaveControl)
  protected
    procedure ThreadProc; virtual; abstract;
  public
    procedure Start; virtual;
  end;

  TProcBuffer = procedure(apData: PAnsiChar) of object;

  TWaveInThread = class(TWaveThread)
  private
    FProcBuffer: TProcBuffer;
    function GetSamples: Longword;
    procedure Proc8 (apData: PAnsiChar);
    procedure Proc16(apData: PAnsiChar);
    procedure Proc24(apData: PAnsiChar);
  protected
    procedure Execute; override;
  public
    constructor Create(aControl: TWaveIn);
  end;

  TOnVolumeChanged = procedure(aValue: LongWord) of object;

  TWaveOut = class(TWaveControl)
  protected
    function GetVolume: WORD;
    procedure SetVolume(aValue: WORD);
  public
    constructor Create(aSetup: IWaveSetup);
    procedure Start;
    property Volume: WORD read GetVolume write SetVolume;
  end;

  TWaveOutThread = class(TWaveThread)
  private
    procedure FillBuffers(aRep: Integer);
    function GetSamples: Longword;
  protected
    procedure Execute; override;
  public
    constructor Create(aControl: TWaveOut);
  end;

implementation

const
  WAVE_FORMAT_DIRECT = $0008;
  WAVE_FORMAT_EXTENSIBLE = $FFFE;
  KSDATAFORMAT_SUBTYPE_PCM: TGUID = '{00000001-0000-0010-8000-00aa00389b71}'; // cbSize = 22
  KSDATAFORMAT_SUBTYPE_IEEE_FLOAT: TGUID = '{00000003-0000-0010-8000-00aa00389b71}'; // cbSize = 22

type
  TA3 = packed array[0..2] of Byte;
  TVolumeMask = packed array[0..1] of Word;

//------------------------------------------------------------------------------

constructor TWaveControl.Create(aSetup: IWaveSetup);
begin
  inherited Create;
  FBuffers:= nil;
  FThread:= nil;
  FSetup:= aSetup;
  Channels:= 2;
end;

destructor TWaveControl.Destroy;
begin
  Stop;
  ReallocMem(FBuffers, 0);
  inherited;
end;

procedure TWaveControl.GetFormat;
begin
  with FFormatExt, Format do begin
    wBitsPerSample:= ((FSetup.Bits + 7) shr 3) shl 3;
    nSamplesPerSec:= FSetup.Sampling;
    nBlockAlign:= (wBitsPerSample div 8) * nChannels;
    nAvgBytesPerSec:= nSamplesPerSec * nBlockAlign;
    Samples:= FSetup.Bits;
    if Samples > 16 then begin
      wFormatTag:= WAVE_FORMAT_EXTENSIBLE;
      if nChannels = 1 then
        ChannelMask:= 1
      else
        ChannelMask:= 3;
      SubFormat:= KSDATAFORMAT_SUBTYPE_PCM;
      cbSize:= 22;
    end else begin
      wFormatTag:= WAVE_FORMAT_PCM;
      cbSize:= 0;
    end;
  end;
end;

procedure TWaveControl.AllocBuffers(aCount: Integer);
var
  i, aBufLength: LongWord;
  apBuffer: PAnsiChar;
begin
  with FFormatExt.Format do begin
    ReallocMem(FBuffers, aCount*nBlockAlign);
    aBufLength:= (aCount div Length(FWHdrs))*nBlockAlign;
    apBuffer:= FBuffers;
    for i:= 0 to High(FWHdrs) - 1 do begin
      FillChar(FWHdrs[i], SizeOf(FWHdrs[i]), 0);
      with FWHdrs[i] do begin
        Hdr.dwBufferLength:= aBufLength;
        Hdr.lpData:= apBuffer;
        Inc(apBuffer, aBufLength);
        Spl:= Hdr.dwBufferLength div nBlockAlign;
      end;
    end;
    FillChar(FWHdrs[High(FWHdrs)], SizeOf(FWHdrs[High(FWHdrs)]), 0);
    with FWHdrs[High(FWHdrs)] do begin
      Hdr.dwBufferLength:= LongWord(aCount)*nBlockAlign - aBufLength*High(FWHdrs);
      Hdr.lpData:= apBuffer;
      Spl:= Hdr.dwBufferLength div nBlockAlign;
    end;
  end;
end;

procedure TWaveControl.Synchronize(Method: TThreadMethod);
begin
  FThread.Synchronize(Method);
end;

function TWaveControl.GetChannels: Integer;
begin
  Result:= FFormatExt.Format.nChannels;
end;

procedure TWaveControl.SetChannels(aValue: Integer);
begin
  if aValue > 1 then
    FFormatExt.Format.nChannels:= 2
  else
    FFormatExt.Format.nChannels:= 1;
end;

function TWaveControl.Running: Boolean;
begin
  Result:= Assigned(FThread);
end;

procedure TWaveControl.Stop;
begin
  if not Running then
    Exit;
  FThread.Free;
  FThread:= nil;
end;

//------------------------------------------------------------------------------

constructor TWaveThread.Create(aControl: TWaveControl);
begin
  inherited Create(True);
  FControl:= aControl;
  FDeviceHD:= 0;
  FSamples:= 0;
  FEvent:= CreateEvent(nil, false, false, nil);
end;

procedure TWaveThread.NextWHdr;
begin
  if FWHdrID < High(FControl.FWHdrs) then
    Inc(FWHdrID)
  else
    FWHdrID:= 0;
  FpWHdr:= @(FControl.FWHdrs[FWHdrID].Hdr);
  FCount:= FControl.FWHdrs[FWHdrID].Spl;
end;

destructor TWaveThread.Destroy;
begin
  Terminate;
  PulseEvent(FEvent);
  WaitFor;
  CloseHandle(FEvent);
  inherited;
end;

//------------------------------------------------------------------------------

constructor TWaveInThread.Create(aControl: TWaveIn);
var
  i: Integer;
begin
  inherited Create(aControl);
  Priority:= tpHigher;
  with FControl do begin
    GetFormat;
    with FFormatExt.Format do begin
      i:= cnBufTime*nSamplesPerSec div 1000;
      SetLength(FBufData, i * nChannels);
    end;
    AllocBuffers(i*Length(FWHdrs));
  end;
  case FControl.FFormatExt.Format.wBitsPerSample of
     8: FProcBuffer:= Proc8;
    16: FProcBuffer:= Proc16;
    24: FProcBuffer:= Proc24;
  end;
  FWHdrID:= High(FControl.FWHdrs);
  NextWHdr;
  Resume;
  if waveInOpen(@FDeviceHD, FControl.FSetup.Input, @(FControl.FFormatExt), FEvent, 0,
    CALLBACK_EVENT or WAVE_FORMAT_DIRECT) <> MMSYSERR_NOERROR then
      Exit;
  for i:= 0 to High(FControl.FWHdrs) do with FControl do begin
    waveInPrepareHeader(FDeviceHD, @FWHdrs[i].Hdr, SizeOf(FWHdrs[0].Hdr));
    waveInAddBuffer(FDeviceHD, @FWHdrs[i].Hdr, SizeOf(FWHdrs[0].Hdr));
  end;
  waveInStart(FDeviceHD);
end;

function TWaveInThread.GetSamples: Longword;
var
  aMMTime: MMTIME;
begin
  aMMTime.wType:= TIME_SAMPLES;
  aMMTime.sample:= 0;
  WaveInGetPosition(FDeviceHD, @aMMTime, SizeOf(aMMTime));
  if aMMTime.sample < FSamples then
    Result:= High(LongWord) - FSamples + aMMTime.sample  + 1
  else
    Result:= aMMTime.sample - FSamples;
end;

procedure TWaveInThread.Proc8(apData: PAnsiChar);
const
  cnNorm = 1 shl 24;
var
  i: Integer;
  aSrc: ^Byte;
  aRes: Integer;
  aDst: ^LongInt;
begin
  i:= Length(FControl.FBufData);
  aSrc:= Pointer(apData);
  aDst:= @(FControl.FBufData[0]);
  while i > 0 do begin
    aRes:= aSrc^ - 128;
    aDst^:= aRes * cnNorm;
    Inc(aSrc);
    Inc(aDst);
    Dec(i);
  end;
end;

procedure TWaveInThread.Proc16(apData: PAnsiChar);
var
  i: Integer;
  aSrc, aRes: ^SmallInt; // change type for different bits
  aDst: ^LongInt;
  aShift: Integer;
begin
  i:= Length(FControl.FBufData);
  aSrc:= Pointer(apData);
  aDst:= @(FControl.FBufData[0]);
  aShift:= 32 - FControl.FFormatExt.Samples;
  while i > 0 do begin
    aRes:= Pointer(aDst);
    aRes^:= aSrc^;
    aDst^:= aDst^ shl aShift;
    Inc(aSrc);
    Inc(aDst);
    Dec(i);
  end;
end;

procedure TWaveInThread.Proc24(apData: PAnsiChar);
var
  i: Integer;
  aSrc, aRes: ^TA3; // change type for different bits
  aDst: ^LongInt;
  aShift: Integer;
begin
  i:= Length(FControl.FBufData);
  aSrc:= Pointer(apData);
  aDst:= @(FControl.FBufData[0]);
  aShift:= 32 - FControl.FFormatExt.Samples;
  while i > 0 do begin
    aRes:= Pointer(aDst);
    aRes^:= aSrc^;
    aDst^:= aDst^ shl aShift;
    Inc(aSrc);
    Inc(aDst);
    Dec(i);
  end;
end;

procedure TWaveInThread.Execute;
var
  aSamples: LongWord;
  i: Integer;
begin
  repeat
    WaitForSingleObject(FEvent, INFINITE);
    if Terminated then begin
      if FDeviceHD <> 0 then begin
        WaveInReset(FDeviceHD);
        for i:= 0 to High(FControl.FWHdrs) do begin
          waveInUnPrepareHeader(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
          NextWHdr;
        end;
        WaveInClose(FDeviceHD);
      end;
      Break;
    end else begin
      aSamples:= GetSamples;
      while aSamples >= FCount do begin
        waveInUnPrepareHeader(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
        FProcBuffer(FpWHdr.lpData);
        TWaveIn(FControl).ThreadProc;
        waveInPrepareHeader(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
        waveInAddBuffer(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
        Dec(aSamples, FCount);
        Inc(FSamples, FCount);
        NextWHdr;
      end;
    end;
  until False;
end;

//------------------------------------------------------------------------------

procedure TWaveIn.Start;
begin
  if Running then
    Exit;
  FThread:= TWaveInThread.Create(Self);
  if FThread.FDeviceHD = 0 then begin
    FThread.Free;
    FThread:= nil;
  end;
end;

//------------------------------------------------------------------------------

constructor TWaveOutThread.Create(aControl: TWaveOut);
var
  i, N: Integer;
begin
  inherited Create(aControl);
  Priority:= tpHigher;
  with FControl do begin
    GetFormat;
    with FFormatExt.Format do
      N:= Length(FWHdrs)*cnBufTime*nSamplesPerSec div 1000;
    i:= N + Length(FBufData) - 1;
    N:= i div Length(FBufData);
    AllocBuffers(N*(Length(FBufData) div FFormatExt.Format.nChannels));
    FillBuffers(N);
  end;
  FWHdrID:= High(FControl.FWHdrs);
  NextWHdr;
  Resume;
  if waveOutOpen(@FDeviceHD, FControl.FSetup.Output, @(FControl.FFormatExt), FEvent, 0,
    CALLBACK_EVENT or WAVE_FORMAT_DIRECT) <> MMSYSERR_NOERROR then
      Exit;
  for i:= 0 to High(FControl.FWHdrs) do with FControl do begin
    waveOutPrepareHeader(FDeviceHD, @FWHdrs[i].Hdr, SizeOf(FWHdrs[0].Hdr));
    waveOutWrite(FDeviceHD, @FWHdrs[i].Hdr, SizeOf(FWHdrs[0].Hdr));
  end;
end;

procedure TWaveOutThread.FillBuffers(aRep: Integer);
var
  i, N: Integer;
  aSrc: ^LongInt;
  aVal, aShift: Integer;
  aDst8, aVal8, aDst: ^Byte;
  aDst16, aVal16: ^SmallInt;
  aDst24, aVal24: ^TA3;
begin
  aSrc:= @(FControl.FBufData[0]);
  N:= Length(FControl.FBufData);
  i:= N;
  case FControl.FFormatExt.Format.wBitsPerSample of
     8: begin
          aVal8:= @aVal;
          aDst8:= FControl.FBuffers;
          while i > 0 do begin
            aVal:= (aSrc^ shr 24) + 128;
            aDst8^:= aVal8^;
            Inc(aSrc);
            Inc(aDst8);
            Dec(i);
          end;
          aDst:= aDst8;
        end;
    16: begin
          aDst16:= FControl.FBuffers;
          aVal16:= @aVal;
          while i > 0 do begin
            aVal:= aSrc^ shr 16;
            aDst16^:= aVal16^;
            Inc(aSrc);
            Inc(aDst16);
            Dec(i);
          end;
          aDst:= Pointer(aDst16);
        end;
    24: begin
          aDst24:= FControl.FBuffers;
          aShift:= 32 - FControl.FFormatExt.Samples;
          aVal24:= @aVal;
          while i > 0 do begin
            aVal:= aSrc^ shr aShift;
            aDst24^:= aVal24^;
            Inc(aSrc);
            Inc(aDst24);
            Dec(i);
          end;
          aDst:= Pointer(aDst24);
        end;
    else
      aDst:= nil;
  end;
  aSrc:= FControl.FBuffers;
  N:= N * FControl.FFormatExt.Format.wBitsPerSample div 8;
  for i:= 2 to aRep do begin
    Move(aSrc^, aDst^, N);
    Inc(aDst, N);
  end;
end;

function TWaveOutThread.GetSamples: Longword;
var
  aMMTime: MMTIME;
begin
  aMMTime.wType:= TIME_SAMPLES;
  aMMTime.sample:= 0;
  WaveOutGetPosition(FDeviceHD, @aMMTime, SizeOf(aMMTime));
  if aMMTime.sample < FSamples then
    Result:= High(LongWord) - FSamples + aMMTime.sample  + 1
  else
    Result:= aMMTime.sample - FSamples;
end;

procedure TWaveOutThread.Execute;
var
  aSamples: LongWord;
  i: Integer;
begin
  repeat
    WaitForSingleObject(FEvent, INFINITE);
    if Terminated then begin
      if FDeviceHD <> 0 then begin
        WaveOutReset(FDeviceHD);
        for i:= 0 to High(FControl.FWHdrs) do begin
          waveOutUnPrepareHeader(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
          NextWHdr;
        end;
        WaveOutClose(FDeviceHD);
      end;
      Break;
    end else begin
      aSamples:= GetSamples;
      while aSamples >= FCount do begin
        waveOutUnPrepareHeader(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
        waveOutPrepareHeader(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
        waveOutWrite(FDeviceHD, FpWHdr, SizeOf(FpWHdr^));
        Dec(aSamples, FCount);
        Inc(FSamples, FCount);
        NextWHdr;
      end;
    end;
  until False;
end;

//------------------------------------------------------------------------------

constructor TWaveOut.Create(aSetup: IWaveSetup);
begin
  inherited Create(aSetup);
end;

procedure TWaveOut.Start;
begin
  if Running then
    Exit;
  FThread:= TWaveOutThread.Create(Self);
  if FThread.FDeviceHD = 0 then begin
    FThread.Free;
    FThread:= nil;
  end;
end;

function TWaveOut.GetVolume: WORD;
var
  apMask: ^TVolumeMask;
  aVolume, aResult: DWORD;
begin
  apMask:= @aVolume;
  waveOutGetVolume(FSetup.Output, @aVolume);
  aResult:= (apMask[0] + apMask[1]) div 2;
  Result:= aResult;
end;

procedure TWaveOut.SetVolume(aValue: WORD);
var
  apMask: ^TVolumeMask;
  aVolume: DWORD;
begin
  apMask:= @aVolume;
  apMask[0]:= aValue;
  apMask[1]:= aValue;
  waveOutSetVolume(FSetup.Output, aVolume);
end;

end.



